open Tensor_loops
open Ids


module T_tile = Tensor_tile(Arch.AVX512)
module Conv = T_tile.Conv_build(struct
    let dim_gen = Dim.fresh_gen ()
    let x_dim = dim_gen ~name:"x" ()
    let y_dim = dim_gen ~name:"y" ()
    let f_dim = dim_gen ~name:"f" ()
    let c_dim = dim_gen ~name:"c" ()
    let w_dim = dim_gen ~name:"w" ()
    let h_dim = dim_gen ~name:"h" ()
    let input = Tensor.(make_join ~name:"I" [|join_dims x_dim w_dim;
                                              join_dims y_dim h_dim;
                                              single c_dim|])
    let params = Tensor.make ~name:"P" [|w_dim;h_dim;c_dim;f_dim|]
    let output_name = "O"
  end)

module MM = T_tile.MM_build( struct
    let dim_gen = Dim.fresh_gen ()

    let i_dim = dim_gen ~name:"i" ()
    let j_dim = dim_gen ~name:"j" ()
    let k_dim = dim_gen ~name:"k" ()

    let a = Tensor.make ~name:"R" [|i_dim; k_dim|]
    let b = Tensor.make ~name:"B" [|k_dim; j_dim|]
    let c_name = "C"
  end)

let weird = let open T_tile in let open Conv in
  [ V f_dim;  U (2, f_dim); T (32, x_dim); T (4, c_dim); R w_dim;  Pack_tens input;
    T (32, f_dim); R c_dim; R y_dim; R x_dim; R f_dim; R h_dim;]

let auguste_tile (* Gen with IOLB %*) =let open T_tile in let open Conv in
  [
    V f_dim;
    U (2, f_dim);
    U (12, y_dim);
    T (32, c_dim);
    Hoist_vars [c_dim];
    T (4, x_dim);
    R w_dim;
    R h_dim;
    R x_dim;
    (* L2 full *)
    R y_dim;
    R c_dim;
  ]

let auguste_sizes = let open Conv in
  [f_dim, 32]

let dull_tile = let open T_tile in let open MM in
  [R i_dim; R j_dim; R k_dim]

let dull_tile_conv2 = let open T_tile in let open Conv in
  [R f_dim; R x_dim; R y_dim; R w_dim; R c_dim; R h_dim]

let pack_tile_conv2 = let open T_tile in let open Conv in
  [V f_dim; U (2, f_dim); T (32, x_dim); R y_dim; Pack_tens input; R w_dim; R
     x_dim; R c_dim; R f_dim; R h_dim]

let permute_tile = let open T_tile in let open MM in
  [V j_dim; U (2, j_dim); T(16, k_dim); Hoist_vars [k_dim]; Pack_tens c; R k_dim;  R i_dim; R j_dim]

let pack_tile = let open T_tile in let open MM in
  [V j_dim; Pack_tens a; R i_dim; R j_dim; R k_dim]

let other_tile = let open T_tile in let open MM in
  [V j_dim; U (2, k_dim); T (4, i_dim); R i_dim; T(8, j_dim); R k_dim; R j_dim]

let motoko_tile = let open T_tile in let open MM in
  [V j_dim; U (2, j_dim); U (6, i_dim); T (64, k_dim); Hoist_vars [k_dim];
   T (192, j_dim);T (192, i_dim); (*Pack_tens a;*) Pack_tens b; R i_dim; R k_dim; R j_dim]

let miracle_tile_3x3 = let open T_tile in let open MM in
  [V j_dim; U (3, j_dim); U (3, i_dim); T (128, k_dim); Hoist_vars [k_dim];
   R  j_dim; R i_dim; (*Pack_tens a; Pack_tens b;*)
   R k_dim; ]

let miracle_tile_4x4 = let open T_tile in let open MM in
  [V j_dim; U (8, j_dim); U (8, i_dim); T (128, k_dim); Hoist_vars [k_dim];
   R  j_dim; R i_dim; (*Pack_tens a; Pack_tens b;*)
   R k_dim; ]

let miracle_tile = let open T_tile in  let open MM in
  [V j_dim; U (2, j_dim); U (6, i_dim); T (128, k_dim); Hoist_vars [k_dim];
   R  j_dim; R i_dim; Pack_tens a; Pack_tens b;
   R k_dim; ]

let () = 
  let loop_code =
    Conv.gen_code ~dim_sizes:auguste_sizes auguste_tile in
  print_endline loop_code
