open Ids
open Exprs
open Instruction

module type Loopnest_t = sig
  module Inst: Inst_t
  type loop =
    | Once of loop list
    | Statement of Inst.t
    | VLoop of Inst.t (* This loop is guaranteed to contain only vector instructions *)
    | Unroll of {comment: string option; dim_index_id: Index.t; start: VarExpr.t ;
                 size: int; increment: ConstExpr.t; body: loop list}
    | Loop of {comment: string option; dim_index_id: Index.t; aux: (Index.t * VarExpr.t) list; start: VarExpr.t ;
               halt: VarExpr.t; increment: ConstExpr.t; body: loop list} [@@deriving show]

  module Rev_loop_graph : sig
    type key [@@deriving show]
    type node [@@deriving show]
    type graph [@@deriving show]

    val top: key
    val is_top: key -> bool

    val empty: graph

    val new_unroll: ?comment:string -> Index.t -> VarExpr.t ->  int -> ConstExpr.t -> node
    val new_seq: ?comment:string -> ?aux:(Index.t * VarExpr.t) list -> Index.t -> VarExpr.t
      -> VarExpr.t -> ConstExpr.t
      -> node

    val insert_block: graph -> Inst.t list -> key * graph
    val modify_block: graph -> key -> (Inst.t list -> Inst.t list) -> graph
    val hook_block:   graph -> key -> Inst.t list -> graph
    val get_block: graph -> key -> Inst.t list

    val set_top_comment: graph -> string -> graph
    val set_aux : graph -> (Index.t * (Index.t * ConstExpr.t) list) list -> graph 
    val enclose: graph -> node  -> key * graph
    val to_loop: graph -> loop
    val embed_before: graph -> graph -> key -> graph
    val embed_after: graph -> graph -> key -> graph
  end
  val unroll: ?st:VarExpr.t -> int -> Dim.t -> ConstExpr.t  -> loop list -> loop
  val unroll_insts_dim: Dim.id -> VarExpr.t -> int -> ConstExpr.t -> Inst.t list -> Inst.t list
  val loop: ?st:ConstExpr.t -> ?halt:VarExpr.t -> ?incr:ConstExpr.t -> Dim.t  -> loop list -> loop
  val gen_code: (Size.id * int) list -> loop -> string
end

module L(Inst: Inst_t) : Loopnest_t with module Inst := Inst
