open Exprs
open Ids
open Instruction

module Legacy(Inst: Inst_t): sig
  module LN: sig type loop end

  type combine = {in_: Tensor.t; out: Tensor.t; map: (Dim.t * Dim.t list) list }

  val map_back_sizes: combine -> Size.id -> ConstExpr.t

  type range = Full | Singl of VarExpr.t | From of VarExpr.t
             | Until of VarExpr.t | Range of VarExpr.t * VarExpr.t
  val visit_range: Tensor.t -> (Dim.t * (range * 'a)) list -> LN.loop
  val visitor_loop: Tensor.t -> LN.loop
end

module type MM_args_t = sig
  val i_dim: Dim.t
  val j_dim: Dim.t
  val k_dim: Dim.t
  val a: Tensor.t
  val b: Tensor.t
  val c_name: string
end

module type Conv_args_t = sig
  val x_dim: Dim.t
  val w_dim: Dim.t
  val y_dim: Dim.t
  val h_dim: Dim.t
  val f_dim: Dim.t
  val c_dim: Dim.t
  val input: Tensor.t
  val params: Tensor.t
  val output_name: string
end

module Tensor_tile(A: Arch.Vec_arch_t): sig
  module Inst: Inst_t with module A = A
  module LN: Loop_nest.Loopnest_t with module Inst := Inst

  type loop_type = U of int * Dim.t
                 | V of Dim.t 
                 | T of int * Dim.t
                 | Pack_tens of Tensor.t
                 | Hoist_vars of Dim.t list
                 | R of Dim.t [@@deriving show]

  type tile_scheme = loop_type list

  val tile_to_string: tile_scheme -> string
  val dim_tile_size: tile_scheme -> Dim.t -> int

  module MM_build(MM_args: MM_args_t) : sig
    include MM_args_t
    val c: Tensor.t
    val gen_code: ?dim_sizes:((Dim.t * int) list) -> tile_scheme -> string
  end

  module Conv_build(Conv_args: Conv_args_t) : sig
    include Conv_args_t
    val output: Tensor.t
    val gen_code: ?dim_sizes:((Dim.t * int) list) -> tile_scheme -> string
  end
end
